unit ICStringLists;

(*	This file is part of the Internet Configuration system and is placed in the public domain for the benefit of all.

	This module provides a bunch of routines for dealing with string lists, ie
	the data type used by the 'STR#' resource.  These are a grotesquely useful
	data structure, and are used heavily by IC.

	Just in case you're wondering, a 'STR#' resource comprises a two byte
	integer, which is the count of the number of strings in the resource,
	followed by the string data, stored as closely packed Pascal strings.
	Obviously it's not easy to access this from a high-level language, which
	is why this module exists.
	
	The code in this module is all contributed by Peter, so it contains a bunch
	of things that I wouldn't do and doesn't really check for errors that well.
	However it all works well, so I didn't want to mess with it.
	
	Do not attempt to understand this module without an understanding
	of the toolbox routine Munger.  See IM I-468 or IM:Text 5-21 and
	IM:Text 5-75..5-77.
*)

interface

	uses
		Types;

	function NewStrList: Handle;
		(* This function creates a new empty string list. *)
	
	function CountStrList (strListH: Handle): integer;
		(* This function returns the count of the number of strings in the string list. *)
	
	function GetIndStrList (strListH: Handle; index: integer): Str255;
		(* This function returns the index'th string in the string list. If index
			is out of bounds, it returns the empty string.
		*)
	
	procedure SetIndStrList (strListH: Handle; index: integer; s: Str255);
		(* This routine sets the index'th string in the string list to s.
			index must be greater than or equal to 1.
			If index is greater than the number of strings in the list,
			empty strings are added to the list to pad it up to the appropriate
			length.
		*)
	
	procedure DeleteIndStrList (strListH: Handle; index: integer);
		(* This routine deletes the index'th string in the string list.
			index must be greater than or equal to 1.
			If index is greater than the number of strings in the list,
			this routine does nothing.
		*)
	
	procedure InsertIndStrList (strListH: Handle; index: integer; s: Str255);
		(* This routine inserts the string s at the index'th position
			in the string list, moving strings beyond that up a notch.
			index must be greater than or equal to 1.
			If index is greater than the number of strings in the list,
			empty strings are added to the list to pad it up to the appropriate
			length.
		*)

	function ValidStrList(strListH : Handle) : Boolean;
		(* This function returns true if strListH is a valid string list. *)

	function GetIndStr (strListID: integer; index: integer): Str255;
		(* This routine is a basic wrapper around GetIndString that returns
			the string as function result, which is much more useful than
			as a var parameter.  Note that this routine is different from
			all the others in that the string list is specified implicitly by
			resource ID, rather than explicitly by a Handle.
		*)
	
implementation

	uses
		Memory,
		Resources,
		ToolUtils;

	type
		IntegerHandle = ^IntegerPtr;
		(* IntegerPtr is defined in Types.p, but not IntegerHandle.  *sigh* *)

	function NewStrList: Handle;
		(* See comment in interface part. *)
	begin
		NewStrList := NewHandleClear(sizeof(integer));
	end; (* NewStrList *)

	function CountStrList (strListH: Handle): integer;
		(* See comment in interface part. *)
	begin
		CountStrList := IntegerHandle(strListH)^^;
	end; (* CountStrList *)

	function GetIndStrList (strListH: Handle; index: integer): Str255;
		(* See comment in interface part. *)
		var
			result: Str255;
			count : integer;
			stringListIndex: integer;
			position: longint;
	begin
		(* See comment in interface part. *)
		count := IntegerHandle(strListH)^^;
		if (1 <= index) and (index <= count) then begin
			position := sizeof(integer);
			for stringListIndex := 1 to index - 1 do begin
				position := position + band(Ptr(ord(strListH^) + position)^, $FF) + 1;
			end; (* for *)
			BlockMoveData(Ptr(ord(strListH^) + position), 
									@result, 
									band(Ptr(ord(strListH^) + position)^, $FF) + 1);
		end else begin
			result := '';
		end; (* if *)
		GetIndStrList := result;
	end; (* GetIndStrList *)

	procedure SetIndStrList (strListH: Handle; index: integer; s: Str255);
		(* See comment in interface part. *)
		var
			count : integer;
			strListSize: longint;
			position : longint;
			i: integer;
			junk : longint;
	begin
		count := IntegerHandle(strListH)^^;
		
		(* If we're setting an element beyond the end of the string list,
			pad up the string list with a bunch of empty strings.
		*)
		strListSize := GetHandleSize(strListH);
		if index > count then begin
			SetHandleSize(strListH, strListSize + index - count);
			for position := ord(strListH^) + strListSize to ord(strListH^) + strListSize + index - count - 1 do
				Ptr(position)^ := 0;
			IntegerHandle(strListH)^^ := index;
			count := index;
		end; (* if *)
		
		(* Now find the index'th string and set it. *)
		position := sizeof(integer);
		for i := 1 to index - 1 do begin
			position := position + band(Ptr(ord(strListH^) + position)^, $FF) + 1;
		end; (* for *)
		junk := Munger(strListH, position, nil, band(Ptr(ord(strListH^) + position)^, $FF) + 1, @s, length(s) + 1);
	end; (* SetIndStrList *)

	procedure DeleteIndStrList (strListH: Handle; index: integer);
		(* See comment in interface part. *)
		var
			count : integer;
			i: integer;
			junk: longint;
			position: longint;
	begin
		count := IntegerHandle(strListH)^^;
		if index <= count then begin
			position := sizeof(integer);
			for i := 1 to index - 1 do begin
				position := position + band(Ptr(ord(strListH^) + position)^, $FF) + 1;
			end; (* for *)
			junk := Munger(strListH, position, nil, band(Ptr(ord(strListH^) + position)^, $FF) + 1, @junk, 0);
			IntegerHandle(strListH)^^ := count - 1;
		end; (* if *)
	end; (* DeleteIndStrList *)

	procedure InsertIndStrList (strListH: Handle; index: integer; s: Str255);
		(* See comment in interface part. *)
		var
			count : integer;
			i: integer;
			err: longint;
			position: longint;
			t: string[2];
	begin
		count := IntegerHandle(strListH)^^;
		if index <= count then begin
			position := sizeof(integer);
			for i := 1 to index - 1 do begin
				position := position + band(Ptr(ord(strListH^) + position)^, $FF) + 1;
			end; (* for *)
			t := '';
			err := Munger(strListH, position, nil, 0, @t, length(t) + 1);
			IntegerHandle(strListH)^^ := count + 1;
		end; (* if *)
		SetIndStrList(strListH, index, s)
	end; (* InsertIndStrList *)

	function ValidStrList(strListH : Handle) : Boolean;
		(* See comment in interface part. *)
		var
			result : Boolean;
			count : integer;
			position : longint;
			strListSize : longint;
			numberOfStringsFound : integer;
	begin
		result := (strListH <> nil) & (strListH^ <> nil) & (GetHandleSize(strListH) >= sizeof(integer));
		if result then begin
			count := IntegerPtr(strListH^)^;
			position := sizeof(integer);
			strListSize := GetHandleSize(strListH);
			numberOfStringsFound := 0;
			
			while result & (position < strListSize) do begin
				position := position + band(Ptr(longint(strListH^) + position)^, $00FF) + 1;
				if position > strListSize then begin
					result := false;
				end else begin
					numberOfStringsFound := numberOfStringsFound + 1;
				end; (* if *)
			end; (* while *)
			
			if result & ((numberOfStringsFound <> count) | (position <> strListSize)) then begin
				result := false;
			end; (* if *)
		end; (* if *)
		ValidStrList := result;
	end; (* ValidStrList *)

	function GetIndStr (strListID: integer; index: integer): Str255;
		(* See comment in interface part. *)
		var
			tmpString: Str255;
	begin
		GetIndString(tmpString, strListID, index);
		GetIndStr := tmpString;
	end; (* GetIndStr *)
	
end. (* ICStringLists *)